package info.evccit.utils;

import java.awt.Dimension;

import java.io.EOFException;
import java.io.File;
import java.io.RandomAccessFile;

public class ImageFileDimensions
{
	/**
	 * Signature bytes that identify a GIF format file.
	 */
	static final byte[] GIF_SIGNATURE = {
		(byte) 'G', (byte) 'I', (byte) 'F'
	};
	
	/**
	 * Signature bytes that identify a JPG format file.
	 */
	static final byte[] JPG_SIGNATURE = {
		(byte) 0xff, (byte) 0xd8, (byte) 0xff
	};
	
	/**
	 * Signature bytes that identify a PNG format file.
	 */
	static final byte[] PNG_SIGNATURE = {
		(byte) 137,
		(byte) 'P', (byte) 'N', (byte) 'G',
		(byte) 0x0d, (byte) 0x0a, (byte) 0x1a, (byte) 0x0a
	};
	
	/**
	 * Signature bytes that identify the header chunk of a PNG file.
	 */
	static final byte[] PNG_IHDR  = {
		(byte) 'I', (byte) 'H', (byte) 'D', (byte) 'R'
	};
		
	/**
	 * Retrieve dimensions of a file containing a PNG, JPG, or GIF image.
	 * If the file is not found, cannot be read, or is badly formed,
	 * return a width and height of zero.
	 * @param fileName A String giving the complete file name (plus path)
	 * @return A Dimension object giving the image's width and height
	 */
	public static Dimension getFileDimensions( String fileName )
	{
		byte buf[] = new byte[8];	// large enough for PNG signature
		Dimension d = new Dimension(0, 0);
		RandomAccessFile file = null;
		try
		{
			file = new RandomAccessFile( new File( fileName) , "r" );
			file.read( buf, 0, 8 );
			if (equalBytes( buf, JPG_SIGNATURE, JPG_SIGNATURE.length ))
			{
				d= getJpgDimensions( file );
			}
			else if (equalBytes( buf, PNG_SIGNATURE, PNG_SIGNATURE.length ))
			{
				d = getPngDimensions( file );
			}
			else if (equalBytes( buf, GIF_SIGNATURE, GIF_SIGNATURE.length ))
			{
				d = getGifDimensions( file );
			}
			file.close();
		}
		catch (Exception e)
		{
			d = new Dimension(0, 0);
		}
		return d;
	}
	
	/**
	 * Retrieve width and height of a PNG file.
	 * The first twelve bytes of a PNG file are the header,
	 * followed by a series of information chunks. The
	 * first chunk header is IHDR, and it contains the
	 * width and height in big-endian form. 
	 * @param file A RandomAccessFile that reads a PNG's data
	 * @return A Dimension object giving the PNG's width and height
	 */
	static Dimension getPngDimensions( RandomAccessFile file )
	{
		byte buf[] = new byte[4];
		int nRead = 0;
		int width = 0;
		int height = 0;
		Dimension d = new Dimension( 0, 0 );

		try {
			file.seek( 12 );
			nRead = file.read( buf, 0, 4 );
			if (nRead == 4 && equalBytes( buf, PNG_IHDR, PNG_IHDR.length ))
			{
				width = file.readInt();
				height = file.readInt();
				d = new Dimension( width, height );
			}
		}
		catch (Exception e)
		{
			// do nothing
		}
		return d;
	}

	/**
	 * Retrieve width and height of a JPG file.
	 * The first two bytes of a JPG file are <code>0xff</code> <code>0xd8</code>,
	 * followed by a series of headers-and-info. Each header
	 * consists of a <code>0xff</code>, a tag byte, and a 2-byte length in
	 * big-endian form (most-significant-byte first). If the
	 * tag byte is <code>0xc0</code>, then the following information is the
	 * width and height, also in big-endian form.
	 * @param file A RandomAccessFile that reads a JPG's data
	 * @return A Dimension object giving the JPG's width and height
	 */
	static Dimension getJpgDimensions( RandomAccessFile file )
	{
		byte buf[] = new byte[12];
		int nRead = 0;
		int len = 0;
		int height = 0;
		int width = 0;
	
		Dimension d = new Dimension(0, 0);
		/*
			read in four bytes (0xff, tag byte, and 2-byte length)
			when the tag byte equals 0xc0, that is the part of the
			file which holds the dimensions
		*/
		try
		{
			file.seek( 2 );
			nRead = file.read( buf, 0, 2 );
			while (buf[0] == (byte) 0xff && buf[1] != (byte) 0xc0 && nRead != -1)
			{
				len = file.readShort();
								
				/* skip forwards to next area.
					we subtract 2 from the length since it includes the length
					bytes, which we've already read. */
				nRead = file.skipBytes( len-2 );
				if (nRead != len-2)
				{
					throw new EOFException( "End of file whle skipping JPG bytes" );
				}
				nRead = file.read( buf, 0, 2 );
			}
			
			if (nRead != -1)
			{
				nRead = file.skipBytes(3);	// skip length and another tag byte
				if (nRead == 3)
				{
					height = file.readShort();
					width = file.readShort();
					d = new Dimension( width, height );
				}
			}
		}
		catch (Exception e)
		{
			// do nothing
		}
		return d;
	}

	/**
	 * Retrieve width and height of a GIF file.
	 * The first six bytes of a GIF file are the letters "GIF",
	 * followed by the version number, "87a" or "89a".
	 * These are followed by the width and height as little-endian
	 * numbers (least-significant-byte first).
	 * @param file A RandomAccessFile that reads a GIF's data
	 * @return A Dimension object giving the GIF's width and height
	 */
	static Dimension getGifDimensions( RandomAccessFile file )
	{
		byte buf[] = new byte[4];
		int nRead;
		int width = 0;
		int height = 0;
		try
		{
			file.seek(6); // skips over "GIF87a" or "GIF89a"
			nRead = file.read(buf, 0, 4);
			if (nRead == 4)
			{
				width = unsignedByte(buf[1]) * 256 + buf[0];
				height = unsignedByte(buf[3]) * 256 + buf[2];
			}
			else
			{
				throw new EOFException( "Hit end of file while reading GIF header" );
			}
		}
		catch (Exception e)
		{
			// do nothing
		}
		return new Dimension( width, height );
	}

	/**
	 * Compare two arrays to see if the first nBytes bytes are equal.
	 * @param a First array of bytes
	 * @param b Second array of bytes
	 * @param nBytes Number of bytes to compare
	 * @return True if first nBytes bytes of the arrays are equal
	 */
	static boolean equalBytes( byte[] a, byte[] b, int nBytes )
	{
		boolean result = true;
		if (a.length >= nBytes && b.length >= nBytes )
		{
			int i = 0;
			while ( i < nBytes && a[i] == b[i] )
			{
				i++;
			}
			result = (i == nBytes);
		}
		else
		{
			result = false;
		}
		return result;
	}

	/**
	 * Return the value of a byte as an unsigned value in range 0..255.
	 * @param b A byte value
	 * @return The unsigned value of the byte
	 */
	static int unsignedByte(byte b)
	{
		return (b < 0) ? b + 256 : b;
	}

}
